#include "common.h"
#include <libsec.h>
#include <plumb.h>
#include "dat.h"

typedef struct {
	int debug;
	int refreshtime;
} Mdir;

static const char* 
slurp(const char* f, char** b, ulong* v)
{
	Dir* d;
	int fd;
	ulong length;
	char* p;
	static char err[128];
	
	fd = open((char*)f, OREAD);
	if (0>fd){
		errstr(err, sizeof(err));
		return err;
	}

	d = dirfstat(fd);
	if (!d){
		close(fd);
		errstr(err, sizeof(err));
		return err;
	}
	length = (ulong)d->length;		// you are screwed if you have a message > 4GB.
	free(d);

	p = emalloc(length);
	if (length != readn(fd, p, length)){
		errstr(err, sizeof(err));
		free(p);
		close(fd);
		return err;
	}
	
	close(fd);
	*b = p;
	*v = length;
	return 0;
}

static const char*
mdirfetch(Mailbox *mb, Message *m)
{
	char sdigest[2*SHA1dlen+1];
	char mpath[1024];
	const char* e;
	ulong len;
	Mdir *mdir;
	int i;

	mdir = mb->aux;
	if (mdir->debug)
		fprint(2, "mdirfetch(%s) ...", m->uidl);

	if (m->start) {
		fprint(2, "already fetched\n");
		return 0;
	}

	snprint(mpath, 1024, "%s/%s", mb->path, m->uidl);
	e = slurp(mpath, &m->start, &len);

	if(e) {
		if (mdir->debug)
			fprint(2, "error %s\n", e);
		return e;
	}

	m->end = m->start+len;
	m->bend = m->rbend = m->end;
	m->header = m->start;

	parse(m, 0, mb, 1);

	// digest headers
	sha1((uchar*)m->start, m->end - m->start, m->digest, nil);
	for(i = 0; i < SHA1dlen; i++)
		sprint(sdigest+2*i, "%2.2ux", m->digest[i]);
	m->sdigest = s_copy(sdigest);

	if (mdir->debug)
		fprint(2, "fetched %d bytes\n", len);
	return nil;
}

static int
Dircmp(const Dir* a, const Dir* b){
	return strcmp(a->name, b->name);
}

// must be [0-9]+(\..*)?
static int
Dirskip(const Dir* a){
	long l;
	char c;

	l = strspn(a->name, "0123456789");
	c = a->name[l];

	if (7>l || ('.' != c && 0 != c))
		return 1;
	return 0;
}

static char*
mdirread(Mdir* mdir, Mailbox* mb, int doplumb)
{
	const char *s;
	int i, nnew;
	Message *m, **ll;
	Dir * D;
	int fd;
	int N;
	int c;
	static char err[128];

	if (mdir->debug)
		fprint(2, "mdirread(doplumb=%d, t=%ld)\n", doplumb, time(0));

	fd = open(mb->path, OREAD);
	if (0>fd){
		errstr(err, sizeof(err));
		return err;
	}

	/*
	 *  a new qid.path means reread the mailbox, while
	 *  a new qid.vers means read any new messages
	 */
	D = dirfstat(fd);
	if(D == nil){
		close(fd);
		errstr(err, sizeof(err));
		return err;
	}
	if(mb->d){
		if(D->qid.path == mb->d->qid.path && D->qid.vers == mb->d->qid.vers){
fprint(2, "\tmdir read thinks qids match\n");
			close(fd);
			free(D);
			return nil;
		}
		free(mb->d);
	}

	snprint(err, sizeof err, "reading '%s' (mdir)", mb->path);
	logmsg(err, nil);

	mb->d = D;
	mb->vers++;
	henter(PATH(0, Qtop), mb->name,
		(Qid){PATH(mb->id, Qmbox), mb->vers, QTDIR}, nil, mb);

	N = dirreadall(fd, &D);
	close(fd);

	if (0 == N)
		return nil;
	if (-1 == N){
		errstr(err, sizeof(err));
		return err;
	}

	qsort(D, N, sizeof(*D), (int(*)(const void*, const void*))Dircmp);

	nnew=0;	
	for(i=0, ll=&mb->root->part; *ll || i<N; ){
		c=-1;
		if (i>=N)
			c=1;
		else if (*ll)
			c = strcmp((*ll)->uidl,D[i].name);
		if (c<0){
			if (Dirskip(D+i)) {
				i++;
				continue;
			}
			// new message
fprint(2,"new: %s (%s)\n", D[i].name, *ll ? (*ll)->uidl : "");
			m = newmessage(mb->root);
			m->mallocd = 1;
			m->inmbox = 1;
			snprint(m->uidl, sizeof(m->uidl), "%s", D[i].name);
			m->atime = m->mtime = time(0);

			if(s = mdirfetch(mb, m)){
				// message disappeared?  unchain
				fprint(2, "deleted → %s (%s)\n", s, m->uidl);
				delmessage(mb, m);
				mb->root->subname--;
				nnew--;
				continue;
			}

			// add to chain, will download soon
			nnew++;
			logmsg("new", m);
			(*ll) = m;
			ll = &m->next;
			if(doplumb)
				mailplumb(mb, m, 0);
			i++;
		} else if (c>0){
fprint(2, "deleted: %s (%s)\n", i<N ? D[i].name : 0, *ll ? (*ll)->uidl : "");
			// deleted message;
			logmsg("deleted", *ll);
			if(doplumb)
				mailplumb(mb, (*ll), 1);
			(*ll)->inmbox = 0;
			(*ll)->deleted = 1;
			ll = &(*ll)->next;
		} else {
fprint(2, "same: %s\n", D[i].name);
			logmsg("duplicate", *ll);
			i++;
			ll = &(*ll)->next;
		}
	}
	free(D);

	logmsg("mbox read", nil);

	if(nnew || mb->vers == 0){
		mb->vers++;
		henter(PATH(0, Qtop), mb->name,
			(Qid){PATH(mb->id, Qmbox), mb->vers, QTDIR}, nil, mb);
	}
	return nil;
}

int
mdirpurgedeleted(Mailbox* mb)
{
	Message *m, *next;
	Mdir* mdir;
	int newdels;
	char mpath[1024];

	mdir = mb->aux;

	if (mdir->debug)
		fprint(2, "mdirpurgedeleted()\n");

	newdels = 0;
	for(m = mb->root->part; m != nil; m = next){
		next = m->next;
		if(m->deleted && m->refs == 0){
			if(m->inmbox) {
				snprint(mpath,1024,"%s/%s", mb->path, m->uidl);
				if (mdir->debug)
					fprint(2, "sysremove: %s\n", mpath);
				if (0 != sysremove(mpath))
					sysfatal("sysremove: %s %r", mpath);
				newdels++;
			}
			delmessage(mb, m);
		}
	}
	if (mdir->debug)
		fprint(2, "newdels: %d\n", newdels);
	return newdels;
}

static void
mdirdelete(Mdir* mdir, Mailbox* mb)
{
	USED(mdir);
	USED(mb);
}

static char*
mdirsync(Mailbox* mb, int doplumb)
{
	Mdir* mdir;
	/*const*/char* rv;

	mdir = mb->aux;
//	if (mdir->debug)
		fprint(2, "mdirsync()\n");

	rv = mdirread(mdir, mb, doplumb);
	if (0 < mdirpurgedeleted(mb))
		mdirdelete(mdir, mb);

	return rv;
}

static char*
mdirctl(Mailbox* mb, int c, char**v)
{
	Mdir* mdir;

	if (c<1)
		return "bad mdir control message";

	mdir = mb->aux;
	if (1 == c && 0 == strcmp(*v, "debug")) {
		mdir->debug = 1;
		return nil;
	}

	if (1 == c && 0 == strcmp(*v, "nodebug")) {
		mdir->debug = 0;
		return nil;
	}
	return 0;
}

static void
mdirclose(Mailbox* mb)
{
	Mdir* mdir;

	mdir = mb->aux;
	free(mdir);
}

char*
mdirmbox(Mailbox *mb, char *path)
{
	Mdir* mdir;
//	static char err[64];
	Dir* d;
	char b[1024];

	if (access(path, AEXIST) <0){
		return Enotme;
	}
	
	d = dirstat(path);
	if (!d){
		return Enotme;
	}
	free(d);

	snprint(b,1024,"%s/new", path);
	d = dirstat(b);
	if (!d){
		return Enotme;
	}
	strcpy(mb->path, b);
	free(d);
	
	mdir = emalloc(sizeof(Mdir));
	mdir->debug=1;
	mdir->refreshtime = 10;
	mb->aux = mdir;
	mb->sync = mdirsync;
	mb->close = mdirclose;
	mb->d = emalloc(sizeof(*mb->d));
	mb->fetch = (char* (*)(Mailbox*, Message*))mdirfetch;
	mb->ctl = mdirctl;

	return nil;
}
